/*
  Copyright (c) 2008 Seneca College
  Licenced under the MIT License (http://www.c3dl.org/index.php/mit-license/)
*/
Model.prototype = new Primitive;

/**
	@class A Model is an object which has a geometric representation composed of 
	vertices, normals and uv coordinates. Models may also  have textures
	applied to them.
*/
function Model()
{
	var loadedTexture	= false;

	// is the model ready to be rendered?
	this.readyToRender = false;
	this.path = null;
	
	this.expandedVertices = new Array();
	this.expandedNormals = null;
	this.expandedUVs = null;
	
	this.buffers = {};
	this.firstTimeRender = true;
	
	this.stayInFrontOfCamera = false;
	
	/**
		<p>
		This method should be called after a model object has been 
		created using:
		</p>

		<p>
			<code>
			var model = new Model();
			</code>
		</p>

		<p>
		null values may be provided for the normals and texcoord arguments,
		however the vertices and faces arrays are required.
		</p>

		@param {Array} vertices Array of the unique vertices of the model.

		@param {Array} [normals] Array of the unique normals of the 
		model, if null lighting will not work correctly for this model.

		@param {Array} [texcoords] Array of texture coordinates of 
		the model. If null, textures cannot be applied to this model.

		@param {Array of Arrays} faces Array of arrays which contains index 
		values which refer to values in the previous arrays.
	*/
	this.init = function(vertices, normals, texcoords, faces)
	{
		var texCoordsPresent = false;
		var normalsPresent = false;
		// At the least, vertices and faces must be present
		if( vertices instanceof Array && faces instanceof Array)
		{
			if( texcoords instanceof Array)
			{
				texCoordsPresent = true;
				this.expandedUVs = new Array();
			}

			if( normals instanceof Array)
			{
				normalsPresent = true;
				this.expandedNormals = new Array();
			}

			for( var i =0; i < faces.length; i++)
			{
				// get vert index
				var vertIndex = faces[i][0];
				var cvert = vertices[vertIndex];
				this.expandedVertices.push(cvert[0]);
				this.expandedVertices.push(cvert[1]);
				this.expandedVertices.push(cvert[2]);
							
				// push on uvs
				if( texcoords instanceof Array)
				{
					var texIndex =  faces[i][1];
					var ctex =  texcoords[texIndex];
					this.expandedUVs.push(ctex[0]);
					this.expandedUVs.push(ctex[1]);
				}
				
				// push on normals
				if( normals instanceof Array)
				{					
					var offset = (texCoordsPresent == false) ? 1: 2;

					var normIndex = faces[i][offset];
					var cnorm = normals[normIndex];
					this.expandedNormals.push(cnorm[0]);
					this.expandedNormals.push(cnorm[1]);
					this.expandedNormals.push(cnorm[2]);
				}
			}
		}
	}

	/**
	*/
	this.initDAE = function(vertices, normals, texcoords)
	{
		this.expandedNormals = normals;
		this.expandedVertices = vertices;
		this.expandedUVs = texcoords;
	}

	/**
		@private

		@param {context} glCanvas3D
	*/
	this.createBuffers = function(glCanvas3D)
	{
		this.buffers.vertex = glCanvas3D.createBuffer(glCanvas3D.STATIC_DRAW, 3, glCanvas3D.FLOAT, this.expandedVertices);

		if( this.expandedUVs )
		{
			this.buffers.tex = glCanvas3D.createBuffer(glCanvas3D.STATIC_DRAW, 2, glCanvas3D.FLOAT, this.expandedUVs);
		}

		if( this.expandedNormals )
		{
			this.buffers.normal = glCanvas3D.createBuffer(glCanvas3D.STATIC_DRAW, 3, glCanvas3D.FLOAT, this.expandedNormals);
		}
	}


	/**
		Update animations for linear velocity and angular velocity.

		@param {float} timeStep

		@returns {boolean} true.
	*/
	var updateScale = makeZeroMatrix();
	var updateM = makeZeroMatrix();
	this.update = function(timeStep, camera)
	{
		if (this.stayInFrontOfCamera)
		{
			//!! All this math is crap because it does't work when the camera rotates
			
			var camUp = camera.getUp();
			var camLeft = camera.getLeft();
			var camDir = camera.getDir();
			this.up = makeVector(-camUp[0], -camUp[1], -camUp[2]);
			this.left = makeVector(-camLeft[0], -camLeft[1], -camLeft[2]);
			this.dir = makeVector(-camDir[0], -camDir[1], -camDir[2]);
			
			var camPos = camera.getPosition();
			//!! These hard-coded values are a hack, need to set them based
			//!! on where we want the floating thing to be in the canvas
			this.pos = makeVector(camPos[0] + 10, camPos[1] - 15, camPos[2] + 10);
			
			var d = camera.getDir();
			d = multiplyVector(d, 10);
			this.pos = addVectors( camera.getPosition(), d);

			//logWarning('camera ' + camPos);
			//logWarning('plane ' + this.pos);
			
			return true;
		}
		
		// if the texture is loaded, it requires and update.
		if (this.loadedTexture == true)
		{
			//this.loadedTexture = false;
			return true;
		}
		
		if (isVectorZero(this.linVel) && isVectorZero(this.angVel))
		{
			// nothing will change
			return false;
		}
		
		if (!isVectorZero(this.linVel))
		{
			// Add a velocity to the position
			var velVec = new Array();
			velVec = this.linVel;
			multiplyVector(velVec, timeStep);
			addVectors(this.pos, velVec, this.pos);
		}
		
		if (!isVectorZero(this.angVel))
		{
			// Apply some rotations to the orientation from the angular velocity
			this.pitch(this.angVel[0] * timeStep);
			this.yaw(this.angVel[1] * timeStep);
			this.roll(this.angVel[2] * timeStep);
		}
		
		// force update.
		return true;
	}

	/**
		Render the model.

		@param {context} glCanvas3D
		@param {Scene} scene

	*/
	this.render = function(glCanvas3D, scene)
	{
		if (glCanvas3D == null)
		{
			logWarning('Model::render() called with a NULL glCanvas3D');	
			return false;
		}
			
		
		// render() is passed a reference to the scene, which gives us a chance
		// to setup somethings we could not eariler since we don't have the scene 
		// in some methods.
		if( this.firstTimeRender) // && this.isDae == false )
		{
			if( this.getTextureName())
			{
				scene.getTextureManager().addTexture(this.getTextureName());
			}

			// if we are using the 1.1 context, we have to setup the buffers
			if( scene.getRenderer() instanceof OpenGLES11)
			{
				this.createBuffers(glCanvas3D);
			}
			// make sure this code in this block isn't run again.
			this.firstTimeRender = false;
		}
		else
		{
			// if the user has changed the texture after we have already 
			// rendered once, we have to add the new texture to the 
			// TextureManager
			
			// don't bother to check if the image is already in the 
			// textureManager, the textureManager does that check already.
			if( this.getTextureName() )
			{
				scene.getTextureManager().addTexture(this.getTextureName());
			}
		}
		
		// world transform of the model						
		var wtMat = new Array(	this.getLeft()[0], this.getLeft()[1], this.getLeft()[2], 0,
								this.getUp()[0],   this.getUp()[1],   this.getUp()[2],   0,
								this.getDirection()[0], this.getDirection()[1], this.getDirection()[2], 0,
								this.getPosition()[0],  this.getPosition()[1],  this.getPosition()[2],  1);

		// scaling is done seperately for now					
		var wtScale = new Array( this.getScale()[0],0,0,0,
								0,this.getScale()[1],0,0,
								0,0,this.getScale()[2],0,
								0,0,0,1);	
		
		// only draw the Model if it is visible .
		if( this.isVisible() )
		{
			if(scene.getRenderer() instanceof OpenGLES20)
			{					
				C3DL.pushMatrix();
					C3DL.multMatrix(wtMat);
					C3DL.multMatrix(wtScale);

					var topMatrix = C3DL.peekMatrix();
				C3DL.popMatrix();	

				var modelMatrix = glCanvas3D.getUniformLocation(scene.getRenderer().sp, "modelMatrix");
				glCanvas3D.uniformMatrix( modelMatrix, topMatrix);

				// ambient lighting
				var ambientLight = scene.getAmbientLight();		

				var lightMatrix = new Array( ambientLight[0],0,0,0,
										0,ambientLight[1],0,0,
										0,0,ambientLight[2],0,
										0,0,0,ambientLight[3]);												

				var ambientLightUniform = glCanvas3D.getUniformLocation(scene.getRenderer().sp, "ambientLight");
				glCanvas3D.uniformMatrix(ambientLightUniform, lightMatrix);
				
				// Texture
				
				//var tex = glCanvas3D.getUniformLocation(scene.getRenderer().sp, "myTex");				
				//glCanvas3D.uniformi(tex, TextureManager.getID(this.getTextureName()));
	
				var ta = glCanvas3D.getAttribLocation(scene.getRenderer().sp, "Texture");
				// Only allow textures if the model has a texture and it also has UVs.
				if(this.getTextureName() && this.expandedUVs && ta != -1 )				
				{
					glCanvas3D.activeTexture(glCanvas3D.TEXTURE0);
					glCanvas3D.enable(glCanvas3D.TEXTURE_2D);
					glCanvas3D.bindTexture(glCanvas3D.TEXTURE_2D, scene.getTextureManager().getID(this.getTextureName()));

					glCanvas3D.vertexAttribPointer(ta, 2, glCanvas3D.FLOAT, this.expandedUVs);
					glCanvas3D.enableVertexAttribArray(ta);	
				}
				else
				{
					glCanvas3D.disableVertexAttribArray(ta);
					glCanvas3D.disable(glCanvas3D.TEXTURE_2D);
					glCanvas3D.bindTexture(glCanvas3D.TEXTURE_2D,-1);
				}			


				// NORMALS	
				var na = glCanvas3D.getAttribLocation(scene.getRenderer().sp, "Normal");
				if(this.expandedNormals && na != -1)
				{
					var N = makeZeroMatrix();				
					
					N = addMatrices(N, wtMat);
					N = transposeMatrix(N);
					N = inverseMatrix(N);
					
					var normalMatrix = glCanvas3D.getUniformLocation(scene.getRenderer().sp, "normalMatrix");
					glCanvas3D.uniformMatrix(normalMatrix, N);

					glCanvas3D.vertexAttribPointer(na, 3, glCanvas3D.FLOAT, this.expandedNormals);
					glCanvas3D.enableVertexAttribArray(na);
				}
				else
				{
					glCanvas3D.disableVertexAttribArray(na);					
				}

				// Vertices
				var va = glCanvas3D.getAttribLocation(scene.getRenderer().sp, "Vertex");
				glCanvas3D.vertexAttribPointer(va, 3, glCanvas3D.FLOAT, this.expandedVertices);
				glCanvas3D.enableVertexAttribArray(va);				
				
				// still trying to get this to work,
				// for now, I had to resort to using drawArrays
				//glCanvas3D.drawElements(glCanvas3D.TRIANGLES, ind.length, ind);
				glCanvas3D.drawArrays(glCanvas3D.TRIANGLES, 0, (this.expandedVertices.length/3));			
			}
			else if (scene.getRenderer() instanceof OpenGLES11)
			{
				// TEXTURES 
				// only bind the texture if this object was actually assigned one
				if(this.getTextureName() && this.buffers.tex)
				{
					glCanvas3D.enable(glCanvas3D.TEXTURE_2D);
					glCanvas3D.bindTexture(glCanvas3D.TEXTURE_2D, scene.getTextureManager().getID(this.getTextureName()));
					glCanvas3D.texCoordPointer(this.buffers.tex);
					glCanvas3D.enableClientState(glCanvas3D.TEXTURE_COORD_ARRAY);
				}
				else
				{
					glCanvas3D.disable(glCanvas3D.TEXTURE_2D);
				}

				// NORMALS
				// only enable the normals if the object actually has them
				if( this.buffers.normal)
				{
					glCanvas3D.normalPointer(this.buffers.normal);
					glCanvas3D.enableClientState(glCanvas3D.NORMAL_ARRAY);
				}
				
				// VERTICES
				glCanvas3D.vertexPointer(this.buffers.vertex);
				glCanvas3D.enableClientState(glCanvas3D.VERTEX_ARRAY);
							
				// Draw
				C3DL.pushMatrix();
					C3DL.multMatrix(wtMat);
					C3DL.multMatrix(wtScale);
					
					glCanvas3D.loadMatrix(C3DL.peekMatrix());
					glCanvas3D.drawArrays(glCanvas3D.TRIANGLES, 0, (this.expandedVertices.length/3));

				C3DL.popMatrix();			
				
				// Turn off Buffers
				glCanvas3D.disableClientState(glCanvas3D.VERTEX_ARRAY);
				glCanvas3D.disableClientState(glCanvas3D.NORMAL_ARRAY);
				glCanvas3D.disableClientState(glCanvas3D.TEXTURE_COORD_ARRAY);	
			}// close draw with 1.1
		}// if visible		
	}// render

	// -------------------------------------------------------
	
	// Getters

	/**
		Get the array of the vertices.

		@returns {Array} The array of the vertices.
	*/
	this.getVertices = function()
	{
		return this.expandedVertices; 
	}
	
	/**
		Get the array of the vertices for the bounding box.

		@returns {Array} The array of the vertices for the bounding box.
	*/
	this.getBoundingBox = function()
	{
		var minX = this.expandedVertices[0];
		var maxX = minX;
		var minY = this.expandedVertices[1];
		var maxY = minY;
		var minZ = this.expandedVertices[2];
		var maxZ = minZ;
	
		for (i = 3; i < this.expandedVertices.length; i++) {
			if (i % 3 == 0) {
				if (this.expandedVertices[i] < minX)
					minX = this.expandedVertices[i];
				
				if (this.expandedVertices[i] > maxX)
					maxX = this.expandedVertices[i];
			}
			else if (i % 3 == 1) {
				if (this.expandedVertices[i] < minY)
					minY = this.expandedVertices[i];
				
				if (this.expandedVertices[i] > maxY)
					maxY = this.expandedVertices[i];
			}
			else {
				if (this.expandedVertices[i] < minZ)
					minZ = this.expandedVertices[i];
				
				if (this.expandedVertices[i] > maxZ)
					maxZ = this.expandedVertices[i];
			}
		}
		//var pos = this.getPosition();
		//alert(pos[0] + ", " + pos[1] + ", " + pos[2]);
		//alert(minX + ", " + maxX + "\n" + minY + ", " + maxY + "\n" + minZ + ", " + maxZ + "\n");
		
		var boxVertices = new Array(minX, minY, minZ,
									minX, maxY, minZ,
									maxX, minY, minZ,
									maxX, maxY, minZ,
									minX, minY, maxZ,
									minX, maxY, maxZ,
									maxX, minY, maxZ,
									maxX, maxY, maxZ);
								
		return boxVertices; 
	}
}
